;;; -*- auto-recompile: t -*-

;; This file is not part of GNU Emacs.
;; 
;; surmulot-widget.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; surmulot-widget.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:
;;-------------
;;             Surmulot widgets (O morphs) 

;; last modified February 9, 2008
;; for Emacs 22
;;; --------------------------------------------------------------------------------

(require 'cl)
(require 'timidity-cfg)

;;; ================================================================================
;;;              widget creation
;;; ================================================================================

(defstruct surmulot-widget
  "A surmulot widget, that is a Squeak Morph connected via TCP/IP to Emacs."
  name
  port
  muo-version 
  (insert-function 'surmulot-widget-default-insert-function)
  (eval-function 'surmulot-widget-default-eval-function)) 

(defcustom surmulot-squeak-vm "$SURMULOTDIR/squeak/vm/squeak.exe"
  "Squeak VM for Surmulot"
  :type 'string
  :group 'spfa-paths
  :group 'squeak)

(defun surmulot-squeak-vm ()
  (expand-file-name (substitute-in-file-name surmulot-squeak-vm)))

(defcustom surmulot-squeak-widget-image "$SURMULOTDIR/squeak/muo.image"
  "Squeak image to start when requesting a widget.
Do not use this variable: use function `surmulot-squeak-widget-image' instead."
  :type 'string
  :group 'spfa-paths
  :group 'squeak)

(defcustom surmulot-squeak-widget-use-latest-image t
  "When not nil, use the latest version of `surmulot-squeak-widget-image'"
  :type 'boolean
  :group 'squeak)

(defun surmulot-squeak-widget-image ()
  "Return the Squeak image to start when requesting a widget."
  (if surmulot-squeak-widget-use-latest-image 
      (squeak-latest-image surmulot-squeak-widget-image)
    (expand-file-name (substitute-in-file-name surmulot-squeak-widget-image))))

(defcustom surmulot-squeak-reserved-ports-start 6000
  "TCP/IP local port numbers for widgets start at this value upward"
  :type 'integer
  :group 'squeak)

(defcustom surmulot-widget-width 650
  "Default width in pixels for a widget."
  :type 'integer
  :group 'squeak)

(defcustom surmulot-widget-height 450
  "Default height in pixels for a widget."
  :type 'integer
  :group 'squeak)

(defun surmulot-make-widget (command &optional width height)
  "Have Squeak evaluate COMMAND, which should return a Morph,
and return a surmulot-widget.
WIDTH and HEIGHT optionally set the Morph geometry \(in pixels)."
  (let ((port (surmulot-new-squeak-widget-port))) 
    (save-window-excursion
      (with-temp-buffer ;;  cause du cd !
        (cd (file-name-directory (surmulot-squeak-vm)))
        (shell-command
         (format "%s %s free doIt \"%s\" &"
                 (shell-quote-argument (surmulot-squeak-vm))
                 (shell-quote-argument (surmulot-squeak-widget-image))
                 (format "MuODeployer makeWidget: (%s) extent: (%s@%s) port: %s" 
                         command
                         (or width surmulot-widget-width) 
			 (or height surmulot-widget-height)
                         port))
         (generate-new-buffer-name " *surmulot widget*"))))
    (loop for n from 1 to 10
	  with ok = nil until ok
	  do (sit-for 1)
	  do (condition-case nil
		 (progn
		   (squeak-tcpip-connect port)
		   (setq ok t))
	       (error nil)))
    (squeak-tcpip-connect port)
    (let ((widget (make-surmulot-widget :port port)))
      (setf (surmulot-widget-muo-version widget)
            (squeak-eval-with-return "MuO version" port))
      widget)))

(defun surmulot-new-squeak-widget-port ()
  (loop for port = surmulot-squeak-reserved-ports-start then (1+ port)
        for used in (sort (surmulot-squeak-active-ports) '<)
        if (> used port) return port
        finally return (1+ port)))

(defun surmulot-squeak-active-ports ()
  (save-excursion
    (loop for b in (buffer-list)
          do (set-buffer b)
          if (processp squeak-tcpip-process)
          collect (squeak-port-for-process squeak-tcpip-process))))

;; (surmulot-make-widget "MuO EnvelopeEditor")

;;; ================================================================================
;;;              high-level functions
;;; ================================================================================

(defvar surmulot-widgets-list '()
  "A non updated list of supposedly active widgets.
Use function `surmulot-widgets-list' to get the actual list.")   

(defun surmulot-widgets-list (&optional no-clean-up)
  "Return an updated list of all currently active widgets.
If NO-CLEAN-UP is not nil, do not remove from variable `surmulot-widgets-list'
the widgets those connection is broken"
  (unless no-clean-up
    (setq surmulot-widgets-list
	  (delete-if-not (lambda (w) 
			   (find-if 
			    (lambda (pc)
			      (and (listp pc)
				   (= (cadr pc) (surmulot-widget-port w))))
			    (mapcar 'process-contact (process-list))))
		     surmulot-widgets-list)))
  surmulot-widgets-list)


;;; ================================================================================
;;;              Squeak side protocol
;;; ================================================================================

(defmacro surmulot-widget-define (symb)
  "Define a function 'surmulot-widget-SYMB used by Squeak to invoque the function
in a widget SYMB slot."
  `(defun ,(intern (concat "surmulot-widget-" (prin1-to-string symb))) (&rest args)
     ,(format 
       "This is a private function intended to be called from Squeak. No public usage.
It is evaluated in the comint buffer associated to a given WIDGET: it calls the function
returned by \(surmulot-widget-%s-function WIDGET) with arguments ARGS.
That function is defined by setting the '%s slot in the surmulot-widget structure."
       symb symb)
     (apply
      (funcall ',(intern (concat "surmulot-widget-" (prin1-to-string symb) "-function")) 
	       (find-if (lambda (w) ;; widget associated to the current comint buffer
			  (= (surmulot-widget-port w)
			     (squeak-port-for-process squeak-tcpip-process)))
			surmulot-widgets-list))
      args)))

(surmulot-widget-define insert)
(surmulot-widget-define eval)
; ... more here later

(defun surmulot-widget-default-insert-function (str)
  (save-excursion
    (set-buffer (other-buffer (current-buffer) t))
    (insert str)))

(defun surmulot-widget-default-eval-function (quoted-form)
  (save-excursion
    (set-buffer (other-buffer (current-buffer) t))
    (eval quoted-form)))


;;; ================================================================================
;;;              high-level functions & Widgets menu
;;; ================================================================================

(defcustom surmulot-widget-specs 
  '(;("Envelope Editor" "MuO EnvelopeEditorWidget")
    ("Envelope Editor" "EnvelopeEditor new")
    ("Score Editor" "CsoundScoreEditor new" )
    ("Phrase Editor" "MusicalPhraseEditor new")
    ("Function Editor" "NFunctionEditor new"))
  "List of registered widgets \(\(label constructor) ... ) 
where label is a string identifying the widget type and constructor is the Smalltalk code to evaluate for creating the widget Morph"
  :type '(repeat (list (string :tag "label") (string :tag "constructor")))
  :group 'squeak)

(defun surmulot-widgets-add (name command)
  (let ((widget (surmulot-make-widget command)))
    (add-to-list 'surmulot-widgets-list widget)
    (setf (surmulot-widget-name widget) name)
    widget))

(defun surmulot-make-widgets-menu ()
  (easy-menu-define surmulot-widgets-menu global-map
    "Squeak widgets"
    `("Widgets"
      ("New widget"
       ,@(mapcar 
          (lambda (spec) `[,(car spec) (surmulot-widgets-add ,(car spec) ,(cadr spec)) t])
          surmulot-widget-specs))
      ("Stored widgets"
       ,@(mapcar 
          (lambda (label) `[,label (surmulot-widget-restore ,label) t])
          (surmulot-buffer-widgets)))
      "--"
      ,@(mapcar 
         (lambda (widget)
           `(,(format "%s (port %d)" (surmulot-widget-name widget) (surmulot-widget-port widget))
	     ["store" (surmulot-widget-store ,widget) (cscsd-buffer-is-a-csd-p)]
	     ["interact" (surmulot-widget-interact ,widget) t]
	     ["clone" (surmulot-widget-clone ,widget) t]
	     "--"
             ["close" (surmulot-widget-close ,widget) t]))
         (surmulot-widgets-list)))))

(defun surmulot-widget-close (widget)
  (squeak-eval-no-return "Smalltalk quitPrimitive" 
			 (surmulot-widget-port widget)))

(defun surmulot-widget-do (widget command)
  (squeak-eval-with-return command (surmulot-widget-port widget)))

(defun surmulot-widget-interact (widget)
  (switch-to-buffer-other-window (squeak-comint-buffer (surmulot-widget-port widget))))

(defun surmulot-widget-clone (widget)
  (let ((name (surmulot-widget-name widget))
	(str (surmulot-widget-do 
	      widget 
	      "Widget biggerSurmulotWidget asEncodedString asOneLine")))
    (surmulot-widgets-add name (format "%s reconstituteObject" str))))

(defun surmulot-widget-store (widget)
  "Store the widget NAME as an XML area in the source buffer"
  (let ((label (read-from-minibuffer "Store widget under label: ")) 
	(name (surmulot-widget-name widget))
	(str (surmulot-widget-do widget "Widget biggerSurmulotWidget asEncodedString asOneLine")))
    (save-restriction
      (widen)
      (set-buffer (or (buffer-base-buffer (current-buffer)) (current-buffer)))      
      (when (or (cscsd-buffer-is-a-csd-p)
		(y-or-n-p "The source buffer does not seem to be a valid CSD. Continue ? "))
	(save-excursion
	  (cscsd-go-after-cssynth)
	  (let (beg end ovl)
	    (insert "\n\n<WIDGET>" label ?\n) 
	    (setq beg (point))
	    (insert name ?\n str)
	    (setq end (point))
	    (insert "</WIDGET>\n")
	    (setq ovl (make-overlay beg end))
	    (overlay-put ovl 'invisible 'csd)
	    (overlay-put ovl 'widget 't)))))))

(defun surmulot-buffer-widgets ()
  (when (cscsd-buffer-is-a-csd-p)
    (save-excursion
      (save-restriction
	(widen)
	(goto-char (point-min))
	(let ((wlist))
	  (while (re-search-forward "<WIDGET>\\(.*\\)$" nil t)
	    (add-to-list 'wlist (match-string 1) t))
	  wlist)))))

(defun surmulot-widget-restore (label)
  (save-excursion
    (save-restriction
      (widen)
      (goto-char (point-min))
      (when (search-forward (concat "<WIDGET>" label) nil t)
	(re-search-forward "^.*$")
	(surmulot-widgets-add 
	 (match-string 0)
	 (format "%s reconstituteObject"
		 (buffer-substring-no-properties 
		  (1+ (point))
		  (- (search-forward "</WIDGET>") 9))))))))

(surmulot-make-widgets-menu)
(add-hook 'activate-menubar-hook 'surmulot-make-widgets-menu) ;; ??
(add-hook 'menu-bar-update-hook 'surmulot-make-widgets-menu)


;;; --------------------------- End  -----------------------------

(provide 'surmulot-widget)

;;; surmulot-widget.el ends here







